#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
Copyright (C) 2020, Quantitative Sustainable Design Group

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''

# %%

import numpy as np

__all__ = ('Decay',)


# %%

#!!! This can be potentially made into a Process
class Decay:
    '''For non-steady state degradation.'''

    # Put these as default class attributes
    _COD_max_decay = None
    _decay_k_COD = None
    _MCF_decay = None
    _max_CH4_emission = None
    _N_max_decay = None
    _decay_k_N = None
    _N2O_EF_decay = None

    @staticmethod
    def allocate_N_removal(tot_red, preferred_N):
        '''
        Allocate the total amount of N removal to NH3 and non-NH3 Components.

        Parameters
        ----------
        tot_red : float
            Total amount of N to be removed.
        preferred_N : float
            Current content of the N that will be removed first.

        Returns
        -------
        N removal: tuple
            Amount of preferred N to be removed, amount of other N to be removed.            

        '''
        if not preferred_N > 0:
            return 0, tot_red
        elif preferred_N > tot_red:
            return tot_red, 0
        else:
            return preferred_N, tot_red-preferred_N  
      
    @staticmethod
    def first_order_decay(k, t, max_decay, t0=0, tot=1):
        r'''
        Calculate first-order degradation loss based on Trimmer et al. [1]_
        
        .. math:: C_0 = tot * max_{decay}
        .. math:: C_{avg} = \frac{C_0}{k*t} * (e^{-k*t_0}-e^{-k*t_f})
        .. math:: loss = C_0 - C_{avg}
        
        
        Parameters
        ----------
        k : float
            Degradation rate constant.
        t0 : float
            Degradation time prior to current process.
        t : float
            Degradation time in current process.
        max_decay : float
            Maximum removal ratio.
        tot : float, optional
            Total degradable amount.
            If set to 1 (default), the return is the relative ratio (i.e., loss/tot).

        Returns
        -------
        loss : float
            Amount lost due to degradation.

        References
        ----------
        .. [1] Trimmer et al., Navigating Multidimensional Social–Ecological System
            Trade-Offs across Sanitation Alternatives in an Urban Informal Settlement.
            Environ. Sci. Technol. 2020, 54 (19), 12641–12653.
            https://doi.org/10.1021/acs.est.0c03296.

        '''

        C0 = tot * max_decay
        Cavg = C0/(k*t) * (np.exp(-k*t0)-np.exp(-k*(t0+t)))
        loss = C0 - Cavg
        return loss


    @property
    def COD_max_decay(self):
        '''[float] Maximum fraction of COD removed during storage given sufficient time.'''
        return self._COD_max_decay
    @COD_max_decay.setter
    def COD_max_decay(self, i):
        self._COD_max_decay = float(i)

    @property
    def decay_k_COD(self):
        '''[float] Rate constant for COD decay.'''
        return self._decay_k_COD
    @decay_k_COD.setter
    def decay_k_COD(self, i):
        self._decay_k_COD = float(i)

    @property
    def MCF_decay(self):
        '''[float] Methane correction factor for COD decay.'''
        return self._MCF_decay
    @MCF_decay.setter
    def MCF_decay(self, i):
        self._MCF_decay = float(i)

    @property
    def max_CH4_emission(self):
        '''[float] Maximum methane emssion as a fraction of degraded COD, [g CH4/g COD].'''
        return self._max_CH4_emission
    @max_CH4_emission.setter
    def max_CH4_emission(self, i):
        self._max_CH4_emission = float(i)

    @property
    def N_max_decay(self):
        '''[float] Maximum fraction of N degraded through denitrification during storage given sufficient time.'''
        return self._N_max_decay
    @N_max_decay.setter
    def N_max_decay(self, i):
        self._N_max_decay = float(i)

    @property
    def decay_k_N(self):
        '''[float] Rate constant for N decay.'''
        return self._decay_k_N
    @decay_k_N.setter
    def decay_k_N(self, i):
        self._decay_k_N = float(i)

    @property
    def N2O_EF_decay(self):
        '''[float] Fraction of N emitted as N2O during decay.'''
        return self._N2O_EF_decay
    @N2O_EF_decay.setter
    def N2O_EF_decay(self, i):
        self._N2O_EF_decay = float(i)










#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
Copyright (C) 2020, Quantitative Sustainable Design Group

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''


# %%

import numpy as np
from .. import SanUnit, Construction
from ._decay import Decay
from ..utils.loading import load_data, data_path

__all__ = ('LiquidTreatmentBed',)

data_path += 'sanunit_data/_liquid_treatment_bed.csv'


class LiquidTreatmentBed(SanUnit, Decay):
    '''
    For secondary treatment of liquid based on Trimmer et al. [1]_
    
    Reference documents
    -------------------
    :ref:`qsdsan.sanunits.Decay <sanunits_Decay>`
    
    Parameters
    ----------
    ins : WasteStream
        Waste for treatment.
    outs : WasteStream
        Treated waste, fugitive CH4, and fugitive N2O.
    if_N2O_emission : bool
        If consider N2O emission from N degradation the process.
        
    References
    ----------
    .. [1] Trimmer et al., Navigating Multidimensional Social–Ecological System
        Trade-Offs across Sanitation Alternatives in an Urban Informal Settlement.
        Environ. Sci. Technol. 2020, 54 (19), 12641–12653.
        https://doi.org/10.1021/acs.est.0c03296.
    
    '''
    
    def __init__(self, ID='', ins=None, outs=(), if_N2O_emission=False, **kwargs):

        SanUnit.__init__(self, ID, ins, outs)
        self.if_N2O_emission = if_N2O_emission
    
        data = load_data(path=data_path)
        for para in data.index:
            value = float(data.loc[para]['expected'])
            setattr(self, '_'+para, value)
        del data
        
        for attr, value in kwargs.items():
            setattr(self, attr, value)
    
    _N_ins = 1
    _N_outs = 3

    def _run(self):
        waste = self.ins[0]
        treated, CH4, N2O = self.outs
        treated.copy_like(self.ins[0])
        CH4.phase = N2O.phase = 'g'
        
        # COD removal
        COD_loss = self.first_order_decay(k=self.decay_k_COD,
                                          t=self.tau/365,
                                          max_decay=self.COD_max_decay)
        COD_loss_tot = COD_loss*waste.COD/1e3*waste.F_vol
        
        treated._COD *= (1-COD_loss)
        #!!! Which assumption is better?
        treated.imass['OtherSS'] *= (1-COD_loss)
        # treated.mass *= (1-COD_loss)
        CH4_prcd = COD_loss_tot*self.MCF_decay*self.max_CH4_emission
        CH4.imass['CH4'] = CH4_prcd

        if self.if_N2O_emission:
            N_loss = self.first_order_decay(k=self.decay_k_N,
                                            t=self.tau/365,
                                            max_decay=self.N_max_decay)
            N_loss_tot = N_loss*waste.TN/1e3*waste.F_vol
            NH3_rmd, NonNH3_rmd = \
                self.allocate_N_removal(N_loss_tot, waste.imass['NH3'])
            treated.imass ['NH3'] = waste.imass['NH3'] - NH3_rmd
            treated.imass['NonNH3'] = waste.imass['NonNH3'] - NonNH3_rmd
            N2O.imass['N2O'] = N_loss_tot*self.N2O_EF_decay*44/28
        else:
            N2O.empty()
    
    _units = {
        'Residence time': 'd',
        'Bed length': 'm',
        'Bed width': 'm',
        'Bed height': 'm',
        'Single bed volume': 'm3'
        }

    def _design(self):
        design = self.design_results
        design['Residence time'] = self.tau
        design['Bed number'] = N = self.N_bed
        design['Bed length'] = L = self.bed_L
        design['Bed width'] = W = self.bed_W
        design['Bed height'] = H = self.bed_H
        design['Single bed volume'] = L*W*H
        
        concrete = N*self.concrete_thickness*(L*W+2*L*H+2*W*H)
        self.construction = (
            Construction(item='Concrete', quantity=concrete, unit='m3'),
            )
        self.add_construction()


    def _cost(self):
        pass


    @property
    def tau(self):
        '''[float] Residence time, [d].'''
        return self._tau
    @tau.setter
    def tau(self, i):
        self._tau = float(i)

    @property
    def N_bed(self):
        '''[int] Number of treatment beds, float will be converted to the smallest integer.'''
        return self._N_bed
    @N_bed.setter
    def N_bed(self, i):
        self._N_bed = int(np.ceil(i))

    @property
    def bed_L(self):
        '''[float] Bed length, [m].'''
        return self._bed_L
    @bed_L.setter
    def bed_L(self, i):
        self._bed_L = float(i)

    @property
    def bed_W(self):
        '''[float] Bed width, [m].'''
        return self._bed_W
    @bed_W.setter
    def bed_W(self, i):
        self._bed_W = float(i)

    @property
    def bed_H(self):
        '''[float] Bed height, [m].'''
        return self._bed_H
    @bed_H.setter
    def bed_H(self, i):
        self._bed_H = float(i)

    @property
    def concrete_thickness(self):
        '''[float] Thickness of the concrete wall.'''
        return self._concrete_thickness
    @concrete_thickness.setter
    def concrete_thickness(self, i):
        self._concrete_thickness = float(i)










#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
Copyright (C) 2020, Quantitative Sustainable Design Group

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''


from weakref import WeakKeyDictionary

__all__ = ('Descriptor', 'NonNegativeFloat', 'NonNgeativeInt', 'Fraction')


# %%

class Descriptor(object):
    '''
    Descriptors can be used to make resusable property logics,
    a great reference can be found online. [1]_
    
    Note
    ----
    Using Descriptors to check value properties is slightly slower than using
    the decorators functions in qsdsan.utils.checkers.
    
    References
    ----------
    .. [1] https://nbviewer.jupyter.org/urls/gist.github.com/ChrisBeaumont/5758381/raw/descriptor_writeup.ipynb
     
    '''
    
    __slots__ = ['default', 'name', 'instance', 'owner']
    
    def __init__(self, default=None, name=None):
        self.default = default
        self.name = name
        self.data = WeakKeyDictionary()
        
    def __get__(self, instance, owner):
        return self.data.get(instance, self.default)


class NonNegativeFloat(Descriptor):
    '''For non-negative floats.'''
    
    def __set__(self, instance, value):
        if value < 0:
            if self.name:
                raise ValueError(f'Value for {self.name} cannot be negative.')
            else:
                raise ValueError('Value cannot be negative.')
        self.data[instance] = float(value)


class NonNgeativeInt(Descriptor):
    '''For non-negative integers.'''
    
    def __set__(self, instance, value):
        if value < 0:
            if self.name:
                raise ValueError(f'Value for {self.name} cannot be negative.')
            else:
                raise ValueError('Value cannot be negative.')
        self.data[instance] = int(value)


class Fraction(Descriptor):
    '''For values in [0,1].'''
    
    def __set__(self, instance, value):
        if not 0 <= value <= 1:
            if self.name:
                raise ValueError(f'Value for {self.name} must be in [0,1].')
            else:
                raise ValueError('Value must be in [0,1].')
        self.data[instance] = float(value)
        



        
        
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
Copyright (C) 2020, Quantitative Sustainable Design Group

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''


# %%

from biosteam import Unit, utils
from . import currency, Construction, Transportation
from .utils.piping import WSIns, WSOuts


__all__ = ('SanUnit',)

@utils.registered(ticket_name='SU')
class SanUnit(Unit, isabstract=True):    

    '''
    Subclass of Unit in biosteam, is initialized with WasteStream rather than Stream.
    
    Reference documents
    -------------------
    `biosteam.Unit <https://biosteam.readthedocs.io/en/latest/Unit.html>`_
    
    Additional attributes
    ---------------------
    construction : tuple
        Contains construction information.
    construction_impacts : dict
        Total impacts associated with this unit.
    transportation : tuple
        Contains construction information.
    add_OPEX : float
        Operating expense per hour in addition to utility cost (assuming 100% uptime).
    uptime_ratio : float
        Uptime of the unit to adjust add_OPEX, should be in [0,1]
        (i.e., a unit that is always operating has an uptime_ratio of 1).
    lifetime : float
        Lifetime of this unit in year.
        It will be used to adjust cost and emission calculation in TEA and LCA.
        If left as None, its lifetime will be assumed to be the same as the
        TEA/LCA lifetime.

    '''
    
    _stacklevel = 7

    def __init__(self, ID='', ins=None, outs=(), thermo=None, **kwargs):
        self._register(ID)
        self._specification = None
        self._load_thermo(thermo)
        self._init_ins(ins)
        self._init_outs(outs)
        self._init_utils()
        self._init_results()
        self._assert_compatible_property_package()
        for attr, val in kwargs.items():
            setattr(self, attr, val)

    def _init_ins(self, ins):
        self._ins = WSIns(self, self._N_ins, ins, self._thermo,
                          self._ins_size_is_fixed, self._stacklevel)
        
    def _init_outs(self, outs):
        self._outs = WSOuts(self, self._N_outs, outs, self._thermo,
                            self._outs_size_is_fixed, self._stacklevel)

    def _init_results(self):
        super()._init_results()
        self._add_OPEX = 0.
        self._uptime_ratio = 1.
        self._lifetime = None
        self._construction = ()
        self._transportation = ()


    def _info(self, T, P, flow, composition, N, _stream_info):
        '''Information of the unit.'''
        if self.ID:
            info = f'{type(self).__name__}: {self.ID}\n'
        else:
            info = f'{type(self).__name__}\n'
        info += 'ins...\n'
        i = 0
        for stream in self.ins:
            if not stream:
                info += f'[{i}] {stream}\n'
                i += 1
                continue
            if _stream_info:
                stream_info = stream._info(T, P, flow, composition, N) + \
                    '\n' + stream._wastestream_info()
            else:
                stream_info = stream._wastestream_info()
            su = stream._source
            index = stream_info.index('\n')
            source_info = f'  from  {type(su).__name__}-{su}\n' if su else '\n'
            info += f'[{i}] {stream.ID}' + source_info + stream_info[index+1:] + '\n'
            i += 1
        info += 'outs...\n'
        i = 0
        for stream in self.outs:
            if not stream:
                info += f'[{i}] {stream}\n'
                i += 1
                continue
            if _stream_info:
                stream_info = stream._info(T, P, flow, composition, N) + \
                    '\n' + stream._wastestream_info()
            else:
                stream_info = stream._wastestream_info()
            su = stream._sink
            index = stream_info.index('\n')
            sink_info = f'  to  {type(su).__name__}-{su}\n' if su else '\n'
            info += f'[{i}] {stream.ID}' + sink_info + stream_info[index+1:] + '\n'
            i += 1
        info = info.replace('\n ', '\n    ')
        return info[:-1]
    
    
    _impact = utils.NotImplementedMethod


    def _summary(self):
        '''After system converges, design the unit and calculate cost and environmental impacts.'''
        self._design()
        self._cost()
        self._impact()
    
    
    def show(self, T=None, P=None, flow='g/hr', composition=None, N=15, stream_info=True):
        """Print information of the unit, including WasteStream-specific information"""
        print(self._info(T, P, flow, composition, N, stream_info))
    
    def add_construction(self, add_unit=True, add_design=True, add_cost=True):
        '''Batch-adding construction unit, designs, and costs.'''
        for i in self.construction:
            if add_unit:
                self._units[i.item.ID] = i.item.functional_unit
            if add_design:
                self.design_results[i.item.ID] = i.quantity
            if add_cost:
                self.purchase_costs[i.item.ID] = i.cost 
    
    @property
    def components(self):
        '''[Components] The ``Components`` object associated with this unit.'''
        return self.chemicals
    
    @property
    def construction(self):
        '''[tuple] Contains construction information.'''
        return self._construction
    @construction.setter
    def construction(self, i):
        if isinstance(i, Construction):
            i = (i,)
        else:
            if not iter(i):
                raise TypeError(f'Only <Construction> can be included, not {type(i).__name__}.')
            for j in i:
                if not isinstance(j, Construction):
                    raise TypeError(f'Only <Construction> can be included, not {type(j).__name__}.')
        self._construction = i

    @property
    def construction_impacts(self):
        '''[dict] Total impacts associated with this unit.'''
        impacts = {}
        if not self.construction:
            return impacts
        for i in self.construction:
            impact = i.impacts
            for i, j in impact.items():
                try: impacts[i] += j
                except: impacts[i] = j
        return impacts

    @property
    def transportation(self):
        '''[tuple] Contains transportation information.'''
        return self._transportation
    @transportation.setter
    def transportation(self, i):
        if isinstance(i, Transportation):
            i = (i,)
        else:
            if not iter(i):
                raise TypeError(f'Only <Transportation> can be included, not {type(i).__name__}.')
            for j in i:
                if not isinstance(j, Transportation):
                    raise TypeError(f'Only <Transportation> can be included, not {type(j).__name__}.')
        self._transportation = i

    @property
    def add_OPEX(self):
        '''[float] Operating expense per hour in addition to utility cost.'''
        return self._add_OPEX
    @add_OPEX.setter
    def add_OPEX(self, i):
        self._add_OPEX = i

    def results(self, with_units=True, include_utilities=True,
                include_total_cost=True, include_installed_cost=False,
                include_zeros=True, external_utilities=(), key_hook=None):
        results = super().results(with_units, include_utilities,
                                  include_total_cost, include_installed_cost,
                                  include_zeros, external_utilities, key_hook)
        if with_units:
            results.loc[('Additional OPEX', ''), :] = ('USD/hr', self.add_OPEX)
            results.replace({'USD': f'{currency}', 'USD/hr': f'{currency}/hr'}, inplace=True)
        else:
            results.loc[('Additional OPEX', ''), :] = self.add_OPEX
        return results

    @property
    def uptime_ratio(self):
        '''
        [float] Uptime of the unit to adjust add_OPEX, should be in [0,1]
        (i.e., a unit that is always operating).
        '''
        return self._uptime_ratio
    @uptime_ratio.setter
    def uptime_ratio(self, i):
        assert 0 <= i <= 1, f'Uptime must be between 0 and 1 (100%), not {i}.'
        self._uptime_ratio = float(i)

    @property
    def lifetime(self):
        '''
        [float] Lifetime of this unit in year.
        It will be used to adjust cost and emission calculation in TEA and LCA.
        If left as None, its lifetime will be assumed to be the same as the
        TEA/LCA lifetime.
        '''
        return self._lifetime
    @lifetime.setter
    def lifetime(self, i):
        self._lifetime = float(i)













#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
Copyright (C) 2020, Quantitative Sustainable Design Group

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''


# %%

__all__ = ('format_number', )


def format_number(number):
    number = float(number)
    if abs(number) == 0 or str(number) == 'nan':
        return '0'
    elif abs(number) > 1e6 or abs(number) < 1e-4:
        return str(f'{number:.2E}')
    elif 1e-3 < abs(number) < 0.01:
        return str(round(number, 4))
    elif number == int(number):
        return str(int(number))
    else:
        return str(round(number, 2))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
Copyright (C) 2020, Quantitative Sustainable Design Group

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''

from warnings import warn
from biosteam.utils.piping import MissingStream, StreamSequence
from .. import WasteStream as WS

__all__ = ('MissingWS', 'WSIns', 'WSOuts')


# %%

def as_ws(ws):
    if isinstance(ws, WS):
        return ws
    elif isinstance(ws, str):
        return WS(ws)
    elif ws is None:
        return MissingWS(None, None)
    

class MissingWS(MissingStream):
    '''
    A subclass of ``MissingStream`` in ``biosteam``, Create a ``MissingWS`` (MissingWasteStream)
    object that acts as a dummy in ``Ins`` and ``Outs`` objects until replaced by an
    actual ``WasteStream`` object
    '''

    def materialize_connection(self, ID=''):
        '''
        Disconnect this ``MissingWS`` from any unit operations and replace
        it with an actual ``WasteStream``.
        '''
        source = self._source
        sink = self._sink
        assert source and sink, (
            "both a source and a sink is required to materialize connection")
        material_stream = WS(ID, thermo=source.thermo)
        source._outs.replace(self, material_stream)
        sink._ins.replace(self, material_stream)

    # TODO: add others
    @property
    def COD(self):
        return 0.
    @property
    def BOD(self):
        return 0.
    @property
    def TC(self):
        return 0.
    @property
    def TOC(self):
        return 0.
    @property
    def TN(self):
        return 0.
    @property
    def TKN(self):
        return 0.
    @property
    def TP(self):
        return 0.
    @property
    def TK(self):
        return 0.
    @property
    def charge(self):
        return 0.

    def __repr__(self):
        return '<MissingWS>'
    
    def __str__(self):
        return 'missing waste stream'
    

# %%


def n_missing(ub, N):
    assert ub >= N, f"size of waste streams exceeds {ub}"
    return ub - N

class WSSequence(StreamSequence):
    '''Abstract class for a sequence of waste streams, subclass of ``StreamSequence`` in ``biosteam``.'''
    
    def __init__(self, size, ws, thermo, fixed_size, stacklevel):
        self._size = size
        self._fixed_size = fixed_size
        dock = self._dock
        redock = self._redock
        if ws == ():
            self._streams = [dock(WS(thermo=thermo)) for i in range(size)]
        else:
            if fixed_size:
                self._initialize_missing_streams()
                if ws:
                    if isinstance(ws, str):
                        self._streams[0] = dock(WS(ws, thermo=thermo))
                    elif isinstance(ws, WS):
                        self._streams[0] = redock(ws, stacklevel)
                    else:
                        N = len(ws)
                        n_missing(size, N) # Assert size is not too big
                        self._streams[:N] = [redock(i, stacklevel+1) if isinstance(i, WS)
                                             else dock(WS(i, thermo=thermo)) for i in ws]
            else:
                if ws:
                    if isinstance(ws, str):
                        self._streams = [dock(WS(ws, thermo=thermo))]
                    elif isinstance(ws, WS):
                        self._streams = [redock(ws, stacklevel)]
                    else:
                        self._streams = [redock(i, stacklevel+1) if isinstance(i, WS)
                                         else dock(WS(i, thermo=thermo))
                                         for i in ws]
                else:
                    self._initialize_missing_streams()

    def _create_missing_stream(self):
        MissingWS(None, None).show()
        return MissingWS(None, None)
    
    def _create_N_missing_streams(self, N):
        return [self._create_missing_stream() for i in range(N)]
    
    def _initialize_missing_streams(self):
        self._streams = self._create_N_missing_streams(self._size)

    def __setitem__(self, index, item):
        if isinstance(index, int):
            if item:
                assert isinstance(item, WS), (
                    f"'{type(self).__name__}' object can only contain "
                    f"'WasteStream' objects; not '{type(item).__name__}'")
            elif not isinstance(item, MissingWS):
                item = self._create_missing_stream()
            self._set_stream(int=index, stream=item, stacklevel=3)
        elif isinstance(index, slice):
            wastestreams = []
            for ws in item:
                if ws:
                    assert isinstance(ws, WS), (
                        f"'{type(self).__name__}' object can only contain "
                        f"'WasteStream' objects; not '{type(ws).__name__}'")
                elif not isinstance(ws, MissingWS):
                    ws = self._create_missing_stream()
                wastestreams.append(ws)
            self._set_streams(slice=index, streams=item, stacklevel=3)
        else:
            raise TypeError("Only intergers and slices are valid "
                           f"indices for '{type(self).__name__}' objects")


    # Below shouldn't be needed for newer biosteam
    def _set_streams(self, slice, streams, stacklevel):
        all_streams = self._streams
        for stream in all_streams[slice]: self._undock(stream)
        all_streams[slice] = streams
        for stream in all_streams:
            self._redock(stream, stacklevel)
        if self._fixed_size:
            size = self._size
            N_streams = len(all_streams)
            if N_streams < size:
                N_missing = n_missing(size, N_streams)
                if N_missing:
                    all_streams[N_streams: size] = self._create_N_missing_streams(N_missing)

    def _set_stream(self, int, stream, stacklevel):
        self._undock(self._streams[int])
        self._redock(stream, stacklevel)
        self._streams[int] = stream


class WSIns(WSSequence):
    '''Create an ``Ins`` object which serves as input streams for a ``Unit`` object.'''
    __slots__ = ('_sink', '_fixed_size')
    
    def __init__(self, sink, size, ws, thermo, fixed_size, stacklevel):
        self._sink = sink
        super().__init__(size, ws, thermo, fixed_size, stacklevel)
    
    @property
    def sink(self):
        return self._sink
    
    def _create_missing_stream(self):
        return MissingWS(None, self._sink)
    
    def _dock(self, ws): 
        ws._sink = self._sink
        return ws

    def _redock(self, ws, stacklevel): 
        sink = ws._sink
        if sink:
            ins = sink._ins
            if ins is not self:
                ins.remove(ws)
                ws._sink = new_sink = self._sink
                if sink._ID and new_sink:
                    warn(f"undocked inlet wastestream {ws} from unit {sink}; "
                         f"{ws} is now docked at {self._sink}", 
                         RuntimeWarning, stacklevel)
        else:
            ws._sink = self._sink
        return ws
    
    def _undock(self, ws): 
        ws._sink = None


class WSOuts(WSSequence):
    '''Create an ``Outs`` object which serves as output streams for a ``Unit`` object.'''
    __slots__ = ('_source',)
    
    def __init__(self, source, size, ws, thermo, fixed_size, stacklevel):
        self._source = source
        super().__init__(size, ws, thermo, fixed_size, stacklevel)
    
    @property
    def source(self):
        return self._source
    
    def _create_missing_stream(self):
        return MissingWS(self._source, None)
    
    def _dock(self, ws): 
        ws._source = self._source
        return ws

    def _redock(self, ws, stacklevel): 
        source = ws._source
        if source:
            outs = source._outs
            if outs is not self:
                # Remove from source
                outs.remove(ws)
                ws._source = new_source = self._source
                if source._ID and new_source:
                    warn(f"undocked outlet wastestream {ws} from unit {source}; "
                         f"{ws} is now docked at {self._source}", 
                         RuntimeWarning, stacklevel)
        else:
            ws._source = self._source
        return ws
    
    def _undock(self, ws): 
        ws._source = None


























#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
Copyright (C) 2020, Quantitative Sustainable Design Group

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''

# %%

from biosteam._graphics import splitter_graphics
from .. import SanUnit

__all__ = ('ComponentSplitter',)


class ComponentSplitter(SanUnit):
    '''
    Split the influent into individual ``Component`` objects,
    the last effluent contains all remaining ``Component`` objects.
    '''
    
    def __init__(self, ID='', ins=None, outs=(), split_keys=()):
        
        SanUnit.__init__(self, ID, ins, outs)
        self.split_keys = split_keys
    
    _ins_size_is_fixed = False
    _outs_size_is_fixed = False
    _graphics = splitter_graphics

    def _run(self):
        last = self.outs[-1]
        last.mix_from(self.ins)
        splitted = []
        num = 0
        for cmps in self.split_keys:
            if isinstance(cmps, str):
                self.outs[num].imass[cmps] = last.imass[cmps]
                last.imass[cmps] = 0
                if cmps in splitted:
                    raise ValueError(f'The Component {cmps} appears more than once in split_dict')
                splitted.append(cmps)
            else:
                try: iter(cmps)
                except:
                    raise ValueError('Elements of the split must be str or iterable, '
                                     f'not {type(cmps).__name__}.')
                for cmp in cmps:
                    self.outs[num].imass[cmp] = last.imass[cmp]
                    last.imass[cmp] = 0
                    if cmp in splitted:
                        raise ValueError(f'The Component {cmps} appears more than once in split_dict')
                    splitted.append(cmp)

            num += 1

    @property
    def split_keys(self):
        '''
        [iterable] An iterable containing IDs of Components to be splitted to
        different effluents. Element of the item in the iterable can be str of
        another iterable containing Component IDs. If the item is also iterable,
        all Components whose ID are in the iterable will be splitted to the same
        effluent. Note that the split is 1 (i.e., all of the Component will be
        diverted to the effluent).
        '''
        return self._split_keys
    @split_keys.setter
    def split_keys(self, i):        
        self._split_keys = i














#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
Copyright (C) 2020, Quantitative Sustainable Design Group

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''


# %%

__all__ = ('FuncGetter',)


class FuncGetter:
    __slots__ = ('func', 'params')
    def __init__(self, func, params):
        self.func = func
        self.params = params

    def __call__(self):
        return self.func(*self.params)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
Copyright (C) 2020, Quantitative Sustainable Design Group

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>
    Joy Cheung <joycheung1994@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''

import numpy as np
import pandas as pd
import thermosteam as tmo
from thermosteam import Chemical, Chemicals, CompiledChemicals
from . import _component

__all__ = ('Components', 'CompiledComponents')

utils = tmo.utils
Component = _component.Component
_component_properties = _component._component_properties
_num_component_properties = _component._num_component_properties
_key_component_properties = _component._key_component_properties
setattr = object.__setattr__
_TMH = tmo.base.thermo_model_handle.ThermoModelHandle
_PH = tmo.base.phase_handle.PhaseHandle


# %%

class UndefinedComponent(AttributeError):
    '''AttributeError regarding undefined Component objects.'''
    def __init__(self, ID):
        super().__init__(repr(ID))

# =============================================================================
# Define the Components class
# =============================================================================

class Components(Chemicals):
    '''
    A subclass of the ``Chemicals`` object in the thermosteam [1]_ package, contains ``Component`` objects as attributes.
    
    Reference documents
    -------------------
    .. [1] `thermosteam.Chemicals <https://thermosteam.readthedocs.io/en/latest/Chemicals.html>`_
    
    '''
    
    def __new__(cls, components, cache=False):
        self = super(Chemicals, cls).__new__(cls)
        isa = isinstance
        setfield = setattr
        for component in components:
            if isa(component, Component):
                setfield(self, component.ID, component)
            else:
                if isa(component, Chemical):
                    raise TypeError(f'{component} is a ``Chemical`` object, use ``Component.from_chemical`` to define a ``Component`` object.')
                raise TypeError(f'Only ``Component`` objects can be included, not a ``{type(component).__name__}`` object.')
        return self
    
    def __setattr__(self, ID, component):
        raise TypeError("Cannot set attribute; use ``{self.ID}.append`` instead.")
    
    def __setitem__(self, ID, component):
        raise TypeError("Cannot set item; use ``{self.ID}.append`` instead.")
    
    def __getitem__(self, key):
        '''Return a ``Component`` object or a list of ``Component`` objects.'''
        dct = self.__dict__
        try:
            if isinstance(key, str):
                return dct[key]
            else:
                return [dct[i] for i in key]
        except KeyError as key_error:
            raise UndefinedComponent(key_error.args[0])
    
    def __contains__(self, component):
        if isinstance(component, str):
            return component in self.__dict__
        elif isinstance(component, Component):
            return component in self.__dict__.values()
        else: # pragma: no cover
            return False        
    
    def append(self, component):
        '''Append a Component'''
        if not isinstance(component, Component):
            if isinstance(component, Chemical):
                raise TypeError(f'{component} is a ``Chemical`` object, use ``Component.from_chemical`` to define a ``Component`` object.')
            else:
                raise TypeError("only ``Component`` objects can be appended, "
                               f"not ``{type(component).__name__}`` object.")
        ID = component.ID
        if ID in self.__dict__:
            raise ValueError(f"{ID} already defined in ``Components``.")
        setattr(self, ID, component)
    
    def extend(self, components):
        '''Extend with more ``Component`` objects.'''
        if isinstance(components, Components):
            self.__dict__.update(components.__dict__)
        else:
            for component in components: self.append(component)
        
    def compile(self):
        '''Cast as a ``CompiledComponents`` object.'''
        setattr(self, '__class__', CompiledComponents)
        CompiledComponents._compile(self)
        
        
    
    _default_data = None
    
    @classmethod
    def load_from_file(cls, path='', use_default_data=False, store_data=False):
        '''
        Create ``Component`` objects based on properties defined in a cvs or an Excel file,
        return a ``Components`` object that contains all created ``Component`` objects.
    
        Parameters
        ----------
        path : str
            File path, the file should end with '.cvs', '.xls', or 'xlsx'
    
        Returns
        -------
        A ``Components`` object that contains all created Component objects.
            
        Note
        ----
        The ``Components`` object needs to be compiled before it is used in simulation
    
        '''
        if use_default_data and cls._default_data is not None:
            data = cls._default_data
        else:
            if path[-4:] == '.csv':
                data = pd.read_csv(path)
            elif path[-4:] == '.xls' or path[-4:] == 'xlsx':
                data = pd.read_excel(path)
            else:
                raise ValueError('Only be csv or Excel files can be used.')
        
        new = cls(())

        for i, cmp in data.iterrows():
            if pd.isna(cmp.measured_as):
                cmp.measured_as = None
            try:
                component = Component(ID = cmp.ID, 
                                      search_ID = str(cmp.CAS), 
                                      measured_as = cmp.measured_as)
            except LookupError:
                try:
                    component = Component(ID = cmp.ID, 
                                          search_ID = 'PubChem='+str(int(cmp.PubChem)),
                                          measured_as = cmp.measured_as) 
                except:
                    if not pd.isna(cmp.formula):
                        component = Component(ID = cmp.ID,
                                              formula = cmp.formula, 
                                              measured_as = cmp.measured_as)            
                    else:
                        component = Component(ID = cmp.ID, 
                                              measured_as = cmp.measured_as)
            for j in _component_properties:
                field = '_' + j
                if pd.isna(cmp[j]): setattr(component, j, None)
                else: setattr(component, field, cmp[j])
            new.append(component)

        if store_data:
            cls._default_data = data
        return new
    
    
    @classmethod
    def load_default(cls, use_default_data=True, sotre_data=True, default_compile=True):
        '''
        Create a ``Components`` object containing default ``Component`` objects.
    
        Parameters
        ----------
        use_default_data : bool, optional
            Whether to use default cache data. The default is True.
        sotre_data : bool, optional
            Whether to store the default data as cache. The default is True.
        default_compile : bool, optional
            Whether to compile the default Components. The default is True.
    
        Returns
        -------
        A ``Components`` or ``CompiledComponents`` object with default ``Component`` objects.
    
        Note
        ----
        [1] Component-specific properties are defined in ./data/component.cvs.

        [2] When default_compile is True, all essential chemical-specific properties
        that are missing will be defaulted to those of water.
    
        '''
        import os
        path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data/_components.csv')
        del os
        new = cls.load_from_file(path=path, use_default_data=True, store_data=True)

        H2O = Component.from_chemical('H2O', tmo.Chemical('H2O'),
                                      i_charge=0, f_BOD5_COD=0, f_uBOD_COD=0,
                                      f_Vmass_Totmass=0, description="Water",
                                      particle_size='Soluble',
                                      degradability='Undegradable', organic=False)
        new.append(H2O)
                
        if default_compile:
            isa = isinstance
            for i in new:
                i.default()
                
                if not i.Tb:
                    if i.particle_size == 'Soluble': i.Tb = tmo.Chemical('urea').Tb
                    elif i.particle_size == 'Dissolved gas': i.Tb = tmo.Chemical('CO2').Tb
                    else: i.Tb = tmo.Chemical('NaCl').Tb
                
                if (isa(i.V, _TMH) and (len(i.V.models)==0 or (i.V[0].Tmin > 298.15 or i.V[0].Tmax < 298.15))) or (isa(i.V, _PH) and (len(i.V.l.models)==0 or (i.V.l[0].Tmin > 298.15 or i.V.l[0].Tmax < 298.15))): 
                    if i.particle_size == 'Soluble': 
                        i.copy_models_from(tmo.Chemical('urea'), names=('V',))                        
                    elif i.particle_size in ('Particulate', 'Colloidal'):
                        try: i.V.add_model(1.2e-5)
                        except AttributeError: 
                            i.V.l.add_model(1.2e-5)    # m^3/mol
                            i.V.s.add_model(1.2e-5)
                    else:
                        i.copy_models_from(tmo.Chemical('CO2'), names=('V',))
                      
                    
                for j in ('sigma', 'epsilon', 'kappa', 'Cn', 'mu', 'Psat', 'Hvap'):
                    if isa(getattr(i, j), _TMH) and len(getattr(i,j).models) > 0: continue
                    elif isa(getattr(i, j), _PH) and len(getattr(i,j).l.models) > 0: continue
                    i.copy_models_from(H2O, names=(j,))
            
            new.compile()
        return new        


# %%

# =============================================================================
# Define the CompiledComponents class
# =============================================================================

chemical_data_array = tmo._chemicals.chemical_data_array

def component_data_array(components, attr):
    data = chemical_data_array(components, attr)
    return data

class CompiledComponents(CompiledChemicals):
    '''
    A subclass of the ``CompiledChemicals`` object in the thermosteam [1]_ package,
    contains ``Component`` objects as attributes.
    
    Reference documents
    -------------------
    .. [1] `thermosteam.CompiledChemicals <https://thermosteam.readthedocs.io/en/latest/Chemicals.html>`_
    
    '''
    
    _cache = {}
    
    def __new__(cls, components, cache=None):
        isa = isinstance
        components = tuple([cmp if isa(cmp, Component) else Component(cmp, cache)
                           for cmp in components])        
        cache = cls._cache
        if components in cache:
            self = cache[components]
        else:
            self = object.__new__(cls)
            setfield = setattr
            for cmp in components:
                setfield(self, cmp.ID, cmp)
            self._compile()
            cache[components] = self
        return self    
    
    def refresh_constants(self):
        '''
        Refresh constant arrays of ``Components`` objects,
        including all chemical and component-specific properties.
        '''
        super().refresh_constants()
        dct = self.__dict__
        components = self.tuple
        for i in _num_component_properties:
            dct[i] = component_data_array(components, i)

    def _compile(self):
        dct = self.__dict__
        tuple_ = tuple # this speeds up the code
        components = tuple_(dct.values())
        CompiledChemicals._compile(self)
        
        for component in components:
            missing_properties = component.get_missing_properties(_key_component_properties)
            if not missing_properties: continue
            missing = utils.repr_listed_values(missing_properties)
            raise RuntimeError(f'{component} is missing key component-related properties ({missing}).')

        for i in _num_component_properties:
            dct[i] = component_data_array(components, i)
        
        dct['s'] = np.asarray([1 if cmp.particle_size == 'Soluble' else 0 for cmp in components])
        dct['c'] = np.asarray([1 if cmp.particle_size == 'Colloidal' else 0 for cmp in components])
        dct['x'] = np.asarray([1 if cmp.particle_size == 'Particulate' else 0 for cmp in components])
        dct['b'] = np.asarray([1 if cmp.degradability != 'Undegradable' else 0 for cmp in components])
        dct['rb'] = np.asarray([1 if cmp.degradability == 'Readily' else 0 for cmp in components])
        dct['org'] = np.asarray([int(cmp.organic) for cmp in components])
    
    def subgroup(self, IDs):
        '''Create a new subgroup of ``Component`` objects.'''
        components = self[IDs]
        new = Components(components)
        new.compile()
        for i in new.IDs:
            for j in self.get_synonyms(i):
                try: new.set_synonym(i, j)
                except: pass
        return new
    
    def index(self, ID):
        '''Return index of specified component.'''
        try: return self._index[ID]
        except KeyError:
            raise UndefinedComponent(ID)

    def indices(self, IDs):
        '''Return indices of multiple components.'''
        try:
            dct = self._index
            return [dct[i] for i in IDs]
        except KeyError as key_error:
            raise UndefinedComponent(key_error.args[0])
    
    def __contains__(self, component):
        if isinstance(component, str):
            return component in self.__dict__
        elif isinstance(component, Component):
            return component in self.tuple
        else: # pragma: no cover
            return False
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
Copyright (C) 2020, Quantitative Sustainable Design Group

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''


# %%

__all__ = ('ureg', 'auom', 'ruom', 'parse_unit')

#!!! Make sure all units are defined in a single registry (same as thermosteam)
#!!! Move Component/WasteStream ones here as well

from thermosteam.units_of_measure import ureg, AbsoluteUnitsOfMeasure, RelativeUnitsOfMeasure
auom = AbsoluteUnitsOfMeasure
ruom = RelativeUnitsOfMeasure

import os
ureg.load_definitions(os.path.dirname(os.path.realpath(__file__)) + '/units_of_measure.txt')
del os


def parse_unit(value):
    str_list = value.split(' ') # for something like 'kg CO2-eq'
    if len(str_list) > 1:
        unit = str_list[0]
        others = ' '.join(str_list.pop(0))
        try: return auom(unit), others
        except: pass
    str_list = value.split('-') # for something like 'MJ-eq'
    if len(str_list) > 1:
        unit = str_list[0]
        others = '-'.join(str_list.pop(0))
        try: return auom(unit), others
        except: pass
    # For something like 'MJ' or 'tonne*km', not at the start as something like 'kg N' will
    # be misinterpreted
    try: return auom(value), ''
    except: pass
    return None, value





# construction_units_of_measure = {
#     'mass': AbsoluteUnitsOfMeasure('kg'),
#     'volume': AbsoluteUnitsOfMeasure('m3'),
#     }


# definitions = uom.definitions
# definitions['GWP'] = 'Global warming potential'




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
Copyright (C) 2020, Quantitative Sustainable Design Group

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''


__all__ = ('NonNegativeFloat', 'NonNgeativeInt', 'Fraction')


'''
Checkers are decorators to check if the given value satisfy a set of constrains,
a great reference can be found online. [1]_

Note
----
Using these decorators are much slower than just adding the checks when defining
them, but probably won't matter since it's still < 1 microseconds

References
----------
.. [1] https://www.programiz.com/python-programming/decorator

'''

def NonNegativeFloat(obj):
    def inner_setter(outer_setter):
        data = outer_setter()
        attr = '_' + outer_setter.__name__
        if data < 0:
            raise ValueError(f'Value for {attr[1:]} cannot be negative.')
        setattr(obj, attr, float(data))
    return inner_setter


def NonNgeativeInt(obj):
    def inner_setter(outer_setter):
        data = outer_setter()
        attr = '_' + outer_setter.__name__
        if data < 0:
            raise ValueError(f'Value for {attr[1:]} cannot be negative.')
        setattr(obj, attr, int(data))
    return inner_setter


def Fraction(obj):
    def inner_setter(outer_setter):
        data = outer_setter()
        attr = '_' + outer_setter.__name__
        if not 0 <=data <=1:
            raise ValueError(f'Value for {attr[1:]} must be between [0, 1].')
        setattr(obj, attr, float(data))
    return inner_setter







#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems
Copyright (C) 2020, Quantitative Sustainable Design Group

This module is developed by:
    Joy Cheung <joycheung1994@gmail.com>
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''

# %%
import numpy as np
import biosteam as bst
from thermosteam import Stream, MultiStream, utils
from . import Components
from ._units_of_measure import auom


__all__ = ('WasteStream',)


_defined_composite_vars = ('COD', 'BOD5', 'BOD', 'uBOD', 'NOD', 'ThOD', 'cnBOD',
                           'C', 'N', 'P', 'K', 'Mg', 'Ca', 'solids', 'charge')

_common_composite_vars = ('_COD', '_BOD', '_uBOD', '_TC', '_TOC', '_TN', 
                          '_TKN', '_TP', '_TK', '_TMg', '_TCa', 
                          '_dry_mass', '_charge', '_ThOD', '_cnBOD')

_ws_specific_slots = (*_common_composite_vars,
                      '_pH', '_SAlk', '_ratios', '_impact_item')

_specific_groups = {'SVFA': ('SAc', 'SProp'),
                    'XStor': ('XOHO_PHA', 'XGAO_PHA', 'XPAO_PHA', 
                              'XGAO_Gly', 'XPAO_Gly'),
                    'XANO': ('XAOO', 'XNOO'),
                    'XBio': ('XOHO', 'XAOO', 'XNOO', 'XAMO', 'XPAO', 
                             'XMEOLO', 'XACO', 'XHMO', 'XPRO', 'XFO'),
                    'SNOx': ('SNO2', 'SNO3'),
                    'XPAO_PP': ('XPAO_PP_Lo', 'XPAO_PP_Hi'),
                    'TKN': ()}

_default_ratios = {'iHi_XPAOPP': 0.5,
                   'iCB_XCB': 0.15,
                   'iBAP_CB': 0.,
                   'iUAP_CB': 0.,
                   'iCUInf_XCUInf': 0.,
                   'iSUInf_SU': 1.,
                   'iXUOHOE_XUE': None,}


vol_unit = auom('L/hr')
conc_unit = auom('mg/L')


# %%

# =============================================================================
# Define the WasteStream class
# =============================================================================

@utils.registered(ticket_name='ws')
class WasteStream(Stream):
    '''
    A subclass of the Stream object in the thermosteam [1]_ package with additional
    attributes and methods for waste treatment.
    
    Reference documents
    -------------------
    .. [1] `thermosteam.Stream <https://thermosteam.readthedocs.io/en/latest/Stream.html>`_
    
    '''
    
    __slots__ = (*Stream.__slots__, *_ws_specific_slots)
    _default_ratios = _default_ratios
    
    def __init__(self, ID='', flow=(), phase='l', T=298.15, P=101325.,
                 units='kg/hr', price=0., thermo=None, 
                 pH=7., SAlk=2.5, COD=None, BOD=None, uBOD=None,
                 TC=None, TOC=None, TN=None, TKN=None, TP=None, TK=None,
                 TMg=None, TCa=None, dry_mass=None, charge=None, ratios=None,
                 ThOD=None, cnBOD=None, impact_item=None, **chemical_flows):
        
        super().__init__(ID=ID, flow=flow, phase=phase, T=T, P=P,
                         units=units, price=price, thermo=thermo, **chemical_flows)
        self._init_ws(pH, SAlk, COD, BOD, uBOD, TC, TOC, TN, TKN,
                      TP, TK, TMg, TCa, ThOD, cnBOD, dry_mass, charge, ratios,
                      impact_item)

    def _init_ws(self, pH=7., SAlk=None, COD=None, BOD=None,
                  uBOD=None, TC=None, TOC=None, TN=None, TKN=None,
                  TP=None, TK=None, TMg=None, TCa=None, ThOD=None, cnBOD=None,
                  dry_mass=None, charge=None, ratios=None, impact_item=None):

        self._pH = pH
        self._SAlk = SAlk
        self._COD = COD
        self._BOD = BOD
        self._uBOD = uBOD
        self._TC = TC
        self._TOC = TOC
        self._TN = TN
        self._TKN = TKN
        self._TP = TP
        self._TK = TK
        self._TMg = TMg
        self._TCa = TCa
        self._ThOD = ThOD
        self._cnBOD = cnBOD
        self._dry_mass = dry_mass
        self._charge = charge
        self._ratios = ratios
        if impact_item:
            impact_item._linked_stream = self
        self._impact_item = impact_item

    
    def show(self, T='K', P='Pa', flow='g/hr', composition=False, N=15,
             stream_info=True, details=True):
        '''
        Print WasteStream information.

        Parameters
        ----------
        T : str, optional
            The unit for temperature. The default is 'K'.
        P : float, optional
            The unit for pressure. The default is 'Pa'.
        flow : str, optional
            The unit for the flow. The default is 'kg/hr'.
        composition : bool, optional
            Whether to show flow information of different Component objects in
            the WasteStream as a percentage. The default is False.
        N : int, optional
            Number of Component objects to print out, when left as None,
            the number depends on the default of thermosteam. The default is 15.
        stream_info : bool, optional
            Whether to print Stream-specific information. The default is True.
        details : bool, optional
            Whether to show the all composite variables of the WasteStream. The default is True.

        '''

        info = ''
        # Stream-related specifications
        if stream_info:
            super().show(T, P, flow, composition, N)
        else:
            info += self._basic_info()
            display_units = self.display_units
            T_units = T or display_units.T
            P_units = P or display_units.P
            info += self._info_phaseTP(self.phase, T_units, P_units)
        info += self._wastestream_info(details=details)
        print(info)
        
    _ipython_display_ = show
    
    
    def _wastestream_info(self, details=True):
        _ws_info = '\n WasteStream-specific properties:'
        # Wastewater-related properties are not relevant for gas or solids
        if self.phase != 'l':
            _ws_info += ' None for non-liquid WasteStreams'
        elif self.F_mass == 0:
            _ws_info += ' None for empty WasteStreams'
        else:
            _ws_info += '\n'
            # Only non-zero properties are shown
            _ws_info += int(bool(self.pH))*f'  pH         : {self.pH:.1f}\n'
            _ws_info += int(bool(self.SAlk))*f'  Alkalinity : {self.SAlk:.1f} mg/L\n'
            if details:
                _ws_info += int(bool(self.COD))   *f'  COD        : {self.COD:.1f} mg/L\n'
                _ws_info += int(bool(self.BOD))   *f'  BOD        : {self.BOD:.1f} mg/L\n'
                _ws_info += int(bool(self.TC))    *f'  TC         : {self.TC:.1f} mg/L\n'
                _ws_info += int(bool(self.TOC))   *f'  TOC        : {self.TOC:.1f} mg/L\n'
                _ws_info += int(bool(self.TN))    *f'  TN         : {self.TN:.1f} mg/L\n'
                _ws_info += int(bool(self.TKN))   *f'  TKN        : {self.TKN:.1f} mg/L\n'
                _ws_info += int(bool(self.TP))    *f'  TP         : {self.TP:.1f} mg/L\n'
                _ws_info += int(bool(self.TK))    *f'  TK         : {self.TK:.1f} mg/L\n'
                # _ws_info += int(bool(self.charge))*f'  charge     : {self.charge:.1f} mmol/L\n'
            else:
                _ws_info += '  ...\n'
            
        return _ws_info

    @property
    def components(self):
        return self.chemicals

    @property
    def ratios(self):
        '''
        The ratios used for estimating WasteStream composition based on user input upon initialization.
        Only meaningful for creating a WasteStream object from scratch.
        If not used or specified, default as None.
        '''
        return self._ratios
    @ratios.setter
    def ratios(self, ratios):
        r = self._ratios or WasteStream._default_ratios
        for name, ratio in ratios.items():
            if name not in r.keys():
                raise ValueError(f"Cannot identify ratio named '{name}'."
                                 f"Must be one of {r.keys()}")
            elif isinstance(ratio, (int, float)) and (ratio > 1 or ratio < 0):
                raise ValueError(f"ratio {name}: {ratio} is out of range [0,1].")
            r[name] = ratio
        self._ratios = r

    def composite(self, variable, subgroup=None, particle_size=None, 
                  degradability=None, organic=None, volatile=None,
                  specification=None):
        
        """
        Calculate any composite variable by specifications.

        Parameters
        ----------
        variable : str
            The composite variable to calculate. One of the followings:
                ('COD', 'BOD5', 'BOD', 'uBOD', 'NOD', 'ThOD', 'cnBOD',
                'C', 'N', 'P', 'K', 'Mg', 'Ca', 
                'solids', 'charge').
        subgroup : CompiledComponents, optional
            A subgroup of CompiledComponents. The default is None.
        particle_size : 'g', 's', 'c', or 'x', optional 
            Dissolved gas ('g'), soluble ('s'), colloidal ('c'), particulate ('x'). 
            The default is None.
        degradability : 'rb', 'sb', 'b' or 'u', optional
            Readily biodegradable ('rb'), slowly biodegradable ('sb'), 
            biodegradable ('b'), or undegradable ('u'). The default is None.
        organic : bool, optional
            Organic (True) or inorganic (False). The default is None.
        volatile : bool, optional
            Volatile (True) or involatile (False). The default is None.
        specification : str, optional
            One of ('SVFA', 'XStor', 'XANO', 'XBio', 'SNOx', 'XPAO_PP', 'TKN'). 
            The default is None.

        Returns
        -------
        value : float
            The estimated value of the composite variable, in [mg/L] or [mmol/L] (for "Charge").

        """
        _get = getattr
        if self.F_vol == 0.:
            return 0.
        
        if variable not in _defined_composite_vars:
            raise KeyError(f"Undefined composite variable {variable},"
                           f"Must be one of {_defined_composite_vars}.")            
        
        #!!! assuming it's a liquid WasteStream
        #TODO: deal with units
        if subgroup:
            cmps = subgroup
            
        else:
            cmps = self.components

        IDs = list(cmps.IDs)            
        if 'H2O' in IDs: IDs.remove('H2O')

        if specification:
            if specification == 'TKN': 
                IDs = [ID for ID in IDs if ID not in ('SN2','SNO2','SNO3')]
            elif specification not in _specific_groups.keys():
                raise KeyError(f"Undefined specification {specification}."
                               f"Must be one of {_specific_groups.keys()}."
                               "Or, try defining 'subgroup'.")
            else: 
                IDs = [ID for ID in IDs if ID in _specific_groups[specification]]
                
        IDs = tuple(IDs)
        cmps = cmps.subgroup(IDs)
        cmp_c = self.imass[IDs]/self.F_vol*1e3      #[mg/L]
        exclude_gas = _get(cmps, 's')+_get(cmps, 'c')+_get(cmps, 'x')        
        
        if variable == 'COD': 
            var = cmps.i_COD * cmp_c * exclude_gas
        elif variable == 'uBOD': 
            var = cmps.i_COD * cmps.f_uBOD_COD * cmp_c * exclude_gas
        elif variable in ('BOD5', 'BOD'): 
            var = cmps.i_COD * cmps.f_BOD5_COD * cmp_c * exclude_gas
        elif variable == 'NOD':
            var = cmps.i_NOD * cmp_c * exclude_gas
        elif variable == 'ThOD':
            var = (cmps.i_NOD + cmps.i_COD) * cmp_c
        elif variable == 'cnBOD':
            var = (cmps.i_NOD + cmps.i_COD * cmps.f_BOD5_COD) * cmp_c * exclude_gas
        elif variable == 'C':
            var = cmps.i_C * cmp_c
        elif variable == 'N':
            var = cmps.i_N * cmp_c * exclude_gas
        elif variable == 'P':
            var = cmps.i_P * cmp_c
        elif variable == 'K':
            var = cmps.i_K * cmp_c
        elif variable == 'Mg':
            var = cmps.i_Mg * cmp_c
        elif variable == 'Ca':
            var = cmps.i_Ca * cmp_c
        elif variable == 'solids':
            var = cmps.i_mass * cmp_c * exclude_gas
            if volatile != None:
                if volatile: var *= cmps.f_Vmass_Totmass
                else: var *= 1-cmps.f_Vmass_Totmass
        else:
            var = cmps.i_charge * cmp_c
        
        dummy = np.ones(len(cmp_c))
        if particle_size:
            if particle_size == 'g': 
                dummy *= 1-exclude_gas
            else:
                dummy *= _get(cmps, particle_size)
        
        if degradability:
            if degradability == 'u': dummy *= 1-_get(cmps, 'b')
            elif degradability == 'b': dummy *= _get(cmps, 'b')
            elif degradability == 'rb': dummy *= _get(cmps, 'rb')
            else: dummy *= _get(cmps, 'b')-_get(cmps, 'rb')

        if organic != None:
            if organic: dummy *= _get(cmps, 'org')
            else: dummy *= 1-_get(cmps, 'org')
        
        return (dummy*var).sum()

    
    @property
    def impact_item(self):
        '''[StreamImpactItem] StreamImpactItem this WasteStream is linked to.'''
        return self._impact_item
    @impact_item.setter
    def impact_item(self, i):
        self._impact_item = i
        i.linked_stream = self

    
    def _liq_sol_properties(self, prop, value):
        if self.phase != 'g':
            return getattr(self, '_'+prop) or value
        else:
            raise AttributeError(f'{self.phase} phase WasteStream does not have {prop}.')
    
    #!!! Add some document
    @property
    def pH(self):
        return self._liq_sol_properties('pH', 7.)

    @property
    def SAlk(self):
        '''[float] Alkalinity in meq/L (or mmol HCO3-/L). Assumed to be mainly bicarbonate.'''
        return self._liq_sol_properties('SAlk', 0.)

    @property
    def COD(self):
        '''[float] Chemical oxygen demand in mg/L.'''
        return self._liq_sol_properties('COD', self.composite('COD'))

    @property    
    def BOD(self):
        '''[float] Biochemical oxygen demand in mg/L. Same as BOD5.'''
        return self._liq_sol_properties('BOD', self.composite('BOD'))

    @property    
    def BOD5(self):
        '''[float] 5-day biochemical oxygen demand, in mg/L. Same as BOD.'''
        return self.BOD

    @property    
    def uBOD(self):
        '''[float] Ultimate biochemical oxygen demand, in mg/L.'''
        return self._liq_sol_properties('uBOD', self.composite('uBOD'))

    @property    
    def cnBOD(self):
        '''[float] Carbonaceous nitrogenous BOD, in mg/L. Biochemical oxygen demand including nitrification.'''
        return self._liq_sol_properties('cnBOD', self.composite('cnBOD'))

    @property    
    def ThOD(self):
        '''[float] Theoretical oxygen demand, in mg/L.'''
        return self._liq_sol_properties('ThOD', self.composite('ThOD'))
    
    #!!! Maybe include C_frac, etc. to calculate C_mass/F_mass - valid for all phases
    # Or a function to calculate it?
    @property
    def TC(self):
        '''[float] Total carbon, in mg/L.'''
        return self._liq_sol_properties('TC', self.composite('C'))
    
    @property
    def TOC(self):
        '''[float] Total organic carbon, in mg/L.'''
        return self._liq_sol_properties('TOC', self.composite('C', organic=True))
        
    @property
    def TN(self):
        '''[float] Total nitrogen, in mg/L.'''
        return self._liq_sol_properties('TN', self.composite('N'))
    
    @property
    def TKN(self):
        '''[float] Total Kjeldahl nitrogen, in mg/L.'''
        return self._liq_sol_properties('TKN', self.composite('N', specification='TKN'))
    
    @property
    def TP(self):
        '''[float] Total phosphorus, in mg/L.'''
        return self._liq_sol_properties('TP', self.composite('P'))
    
    @property
    def TK(self):
        '''[float] Total potassium, in mg/L.'''
        return self._liq_sol_properties('TK', self.composite('K'))
    
    @property
    def TMg(self):
        '''[float] Total magnesium, in mg/L.'''
        return self._liq_sol_properties('TMg', self.composite('Mg'))
    
    @property
    def TCa(self):
        '''[float] Total calcium, in mg/L.'''
        return self._liq_sol_properties('TCa', self.composite('Ca'))
    
    @property
    def dry_mass(self):
        '''[float] Total solids, dry mass of dissolved and suspended solids, in mg/L.'''
        return self._liq_sol_properties('solids', self.composite('solids'))
    
    # TODO: calibrate Charge when weak acids are involved
    # @property
    # def charge(self):
    #     return self._liq_sol_properties('charge', self.composite('charge'))
    

    def copy(self, ID=None):
        new = super().copy()
        new._init_ws()
        for slot in _ws_specific_slots:
            value = getattr(self, slot)
            if slot == '_impact_item' and value:
                value.copy(stream=new)
            else:
                setattr(new, slot, utils.copy_maybe(value))
        return new
    __copy__ = copy

    def copy_like(self, other):
        Stream.copy_like(self, other)
        for slot in _ws_specific_slots:
            value = getattr(other, slot)
            setattr(self, slot, value)
    
    def copy_flow(self, other, IDs=..., *, remove=False, exclude=False, if_copy_ws=False):
        #!!! How to inherit the Stream copy_flow function?
        # Stream.copy_flow(self, other, IDs, remove, exclude)

        chemicals = self.chemicals
        mol = other.mol
        if exclude:
            IDs = chemicals.get_index(IDs)
            index = np.ones(chemicals.size, dtype=bool)
            index[IDs] = False
        else:
            index = chemicals.get_index(IDs)
        
        self.mol[index] = mol[index]
        if remove: 
            if isinstance(other, MultiStream):
                other.imol.data[:, index] = 0
            else:
                mol[index] = 0

        if if_copy_ws:
            for slot in _ws_specific_slots:
                value = getattr(other, slot)
                setattr(self, slot, value)

    def mix_from(self, others):
        Stream.mix_from(self, others)
        for slot in _ws_specific_slots:
            #!!! This need reviewing, might not be good to calculate some
            # attributes like pH
            try: tot = sum(float(getattr(i, slot))*i.F_vol for i in others)
            except: continue
            if tot == 0.:
                setattr(self, slot, None)
            else:
                setattr(self, slot, tot/self.F_vol)


    def get_TDS(self, include_colloidal=True):
        '''
        Total dissolved solids (TDS).

        Parameters
        ----------
        include_colloidal : bool, optional
            Whether to include colloidal components as TDS. The default is True.

        Returns
        -------
        TDS : float
            In mg/L.

        '''
        TDS = self.composite('solids', particle_size='s')
        if include_colloidal:
            TDS += self.composite('solids', particle_size='c')
        return TDS
    
    def get_TSS(self, include_colloidal=False):
        '''
        Total suspended solids (TSS).

        Parameters
        ----------
        include_colloidal : bool, optional
            Whether to include colloidal components as TSS. The default is False.

        Returns
        -------
        TSS : float
            In mg/L.

        '''
        TSS = self.composite('solids', particle_size='x')
        if include_colloidal:
            TSS += self.composite('solids', particle_size='c')        
        return TSS
    
    def get_VSS(self, include_colloidal=False):
        '''[float] Volatile suspended solids, in mg/L.'''
        VSS = self.composite('solids', particle_size='x', volatile=True)
        if include_colloidal:
            VSS += self.composite('solids', particle_size='c', volatile=True)        
        return VSS        
    
    def get_ISS(self):
        '''[float] Inorganic/involatile suspended solids, in mg/L.'''
        return self.composite('solids', particle_size='x', volatile=False)


    @classmethod
    def codstates_inf_model(cls, ID, flow_tot=0., units = ('L/hr', 'mg/L'), 
                            phase='l', T=298.15, P=101325., price=0., thermo=None, 
                            pH=7., SAlk=10., ratios=None, 
                            COD=430., TKN=40., TP=10., iVSS_TSS=0.75, iSNH_STKN=0.9,
                            SNH4=25., SNO2=0., SNO3=0., SPO4=8., 
                            SCa=140., SMg=50., SK=28., SCAT=3., SAN=12., SN2=18., 
                            frSUInf=0.05, frSF=0.2, frXCUInf=0.13, 
                            frSUE=0., frSCH3OH=0., frSAc=0., frSProp=0., 
                            frXOHO=0., frXAOO=0., frXNOO=0., frXAMO=0., frXPAO=0., 
                            frXPRO=0., frXACO=0., frXHMO=0., frXMEOLO=0., frXFO=0.,
                            frXOHO_PHA=0., frXGAO_PHA=0., frXPAO_PHA=0., 
                            frXGAO_Gly=0., frXPAO_Gly=0., frXU_OHO_E=0., frXU_PAO_E=0.,
                            XFePO4=0., XAlPO4=0., XFeOH=0., XAlOH=0., 
                            XMAP=0., XHAP=0., XHDP=0., XPAO_PP=0., 
                            XMgCO3=0., XCaCO3=0., DO=0., SH2=0., SCH4=0.):
        
           
        cmps = Components.load_default(default_compile=True)
        bst.settings.set_thermo(cmps)
        
        cmp_dct = dict.fromkeys(cmps.IDs, 0.)

        new = cls(ID=ID, phase=phase, T=T, P=P, units='kg/hr', price=price, 
                  thermo=thermo, pH=pH, SAlk=SAlk)

        if ratios: new.ratios = ratios
        else: new.ratios = WasteStream._default_ratios
        r = new._ratios

        #************ user-defined states **************        
        cmp_dct['SH2'] = SH2
        cmp_dct['SCH4'] = SCH4
        cmp_dct['SN2'] = SN2
        cmp_dct['SO2'] = DO
        cmp_dct['SNH4'] = SNH4
        cmp_dct['SNO2'] = SNO2
        cmp_dct['SNO3'] = SNO3
        cmp_dct['SPO4'] = SPO4
        cmp_dct['SCO3'] = SAlk * 12 * conc_unit.conversion_factor(units[1])             # 1 meq/L SAlk ~ 1 mmol/L HCO3- ~ 12 mg C/L (12 mg C/mmol HCO3-)
        cmp_dct['SCa'] = SCa
        cmp_dct['SMg'] = SMg
        cmp_dct['SK'] = SK
        cmp_dct['XMAP'] = XMAP
        cmp_dct['XHAP'] = XHAP
        cmp_dct['XHDP'] = XHDP
        cmp_dct['XFePO4'] = XFePO4
        cmp_dct['XAlPO4'] = XAlPO4
        cmp_dct['XFeOH'] = XFeOH
        cmp_dct['XAlOH'] = XAlOH
        cmp_dct['XMgCO3'] = XMgCO3
        cmp_dct['XCaCO3'] = XCaCO3
        cmp_dct['SCAT'] = SCAT
        cmp_dct['SAN'] = SAN
        
        #************ organic components **************
        cmp_dct['SCH3OH'] = COD * frSCH3OH
        cmp_dct['SAc'] = COD * frSAc
        cmp_dct['SProp'] = COD * frSProp
        cmp_dct['SF'] = COD * frSF
        cmp_dct['SU_Inf'] = COD * frSUInf
        cmp_dct['SU_E'] = COD * frSUE
        
        SOrg = sum([v for k,v in cmp_dct.items() if k in ('SCH3OH','SAc','SProp','SF','SU_Inf','SU_E')])
        
        XCU_Inf = COD * frXCUInf
        cmp_dct['CU_Inf'] = XCU_Inf * r['iCUInf_XCUInf']
        cmp_dct['XU_Inf'] = XCU_Inf * (1 - r['iCUInf_XCUInf'])
        
        cmp_dct['XOHO'] = COD * frXOHO
        cmp_dct['XAOO'] = COD * frXAOO
        cmp_dct['XNOO'] = COD * frXNOO
        cmp_dct['XAMO'] = COD * frXAMO
        cmp_dct['XPAO'] = COD * frXPAO
        cmp_dct['XACO'] = COD * frXACO
        cmp_dct['XHMO'] = COD * frXHMO
        cmp_dct['XPRO'] = COD * frXPRO
        cmp_dct['XMEOLO'] = COD * frXMEOLO
        cmp_dct['XFO'] = COD * frXFO
        
        XBio = sum([v for k,v in cmp_dct.items() if k.startswith('x') and k.endswith('O')])
        
        cmp_dct['XOHO_PHA'] = COD * frXOHO_PHA
        cmp_dct['XGAO_PHA'] = COD * frXGAO_PHA
        cmp_dct['XPAO_PHA'] = COD * frXPAO_PHA
        cmp_dct['XGAO_Gly'] = COD * frXGAO_Gly
        cmp_dct['XPAO_Gly'] = COD * frXPAO_Gly
        
        XStor = sum([v for k,v in cmp_dct.items() if k.endswith(('PHA','Gly'))])
        
        cmp_dct['XU_OHO_E'] = COD * frXU_OHO_E
        cmp_dct['XU_PAO_E'] = COD * frXU_PAO_E
        
        XU_E = cmp_dct['XU_OHO_E'] + cmp_dct['XU_PAO_E']
        
        XCB = COD - SOrg - XCU_Inf - XU_E
        CB = XCB * r['iCB_XCB']
        cmp_dct['CB_BAP'] = CB * r['iBAP_CB']
        cmp_dct['CB_UAP'] = CB * r['iUAP_CB']
        cmp_dct['CB_Subst'] = CB - cmp_dct['CB_BAP'] - cmp_dct['CB_UAP']
        
        cmp_dct['XB_Subst'] = XCB - CB - XBio - XStor
        
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        VSS = (cmp_c * cmps.i_mass * cmps.f_Vmass_Totmass * cmps.x * cmps.org).sum()
        TSS = VSS/iVSS_TSS
        XOrg_ISS = (cmp_c * cmps.i_mass * (1-cmps.f_Vmass_Totmass) * cmps.x * cmps.org).sum()

        del SOrg, XCU_Inf, XBio, XStor, XU_E, XCB, CB
        
        #************ inorganic components **************
        cmp_dct['XPAO_PP_Hi'] = XPAO_PP * r['iHi_XPAOPP']
        cmp_dct['XPAO_PP_Lo'] = XPAO_PP * (1 - r['iHi_XPAOPP'])
        
        ISS = TSS - VSS
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        other_ig_iss = (cmp_c * cmps.i_mass * cmps.x * (1-cmps.org)).sum()
        cmp_dct['XIg_ISS'] = ISS - XOrg_ISS - other_ig_iss
        
        del ISS, VSS, TSS, XOrg_ISS, other_ig_iss, cmp_c
        
        # TODO: calibrate pH, SAlk, SCAT, SAN
        bad_vars = {k:v for k,v in cmp_dct.items() if v<0}
        if len(bad_vars) > 0:
            raise ValueError(f"The following state variable(s) was found negative: {bad_vars}.")
        
        del bad_vars
        
        #************ calibrate XB_subst, SF's N, P content *************
        if SNH4 > 0 and cmp_dct['SF'] > 0:
            SFi_N = _calib_SF_iN(cmps, cmp_dct, SNH4/iSNH_STKN)
        XB_Substi_N = _calib_XBsub_iN(cmps, cmp_dct, TKN - SNH4/iSNH_STKN)
        XB_Substi_P = _calib_XBsub_iP(cmps, cmp_dct, TP)
        
        #************ convert concentrations to flow rates *************
        flow_tot /= vol_unit.conversion_factor(units[0])
        factor = conc_unit.conversion_factor(units[1])
            
        cmp_dct = {k:v/factor*flow_tot*1e-6 for k,v in cmp_dct.items()}       # [mg/L]*[L/hr]*1e-6[kg/mg] = [kg/hr]
        dwt = sum(cmp_dct.values())         # dry weight
        
        den = 1
        i = 0
        while True:
            den0 = den
            cmp_dct['H2O'] = flow_tot*den0 - dwt
            new = cls(ID=ID, phase=phase, T=T, P=P, units='kg/hr', price=price, 
                      thermo=thermo, pH=pH, SAlk=SAlk, **cmp_dct)
            den = flow_tot*den0/(new.F_vol*1e3)            
            i += 1
            if abs(den-den0) <= 1e-3: break
            if i > 50: raise ValueError('Density calculation failed to converge within 50 iterations.')
        
        new.components.SF.i_N = SFi_N
        new.components.XB_Subst.i_N = XB_Substi_N
        new.components.XB_Subst.i_P = XB_Substi_P
        
        return new


    @classmethod
    def codbased_inf_model(cls, ID, flow_tot=0., units = ('L/hr', 'mg/L'), 
                           phase='l', T=298.15, P=101325., price=0., thermo=None, 
                           pH=7., SAlk=10., ratios=None, 
                           COD=430., TKN=40., TP=10., iVSS_TSS=0.75, iSNH_STKN=0.9,
                           iSCOD_COD=0.25, iSBOD_SCOD=0.50, iBOD_COD=0.58,
                           SNH4=25., SNO2=0., SNO3=0., SPO4=8., 
                           SCa=140., SMg=50., SK=28., SCAT=3., SAN=12., SN2=18., 
                           CB=40., SCH3OH=0., SAc=0., SProp=0., 
                           XOHO_PHA=0., XGAO_PHA=0., XPAO_PHA=0., 
                           XGAO_Gly=0., XPAO_Gly=0., XU_OHO_E=0., XU_PAO_E=0.,
                           XOHO=0., XAOO=0., XNOO=0., XAMO=0., XPAO=0., 
                           XPRO=0., XACO=0., XHMO=0., XMEOLO=0., XFO=0.,
                           XFePO4=0., XAlPO4=0., XFeOH=0., XAlOH=0., 
                           XMAP=0., XHAP=0., XHDP=0., XPAO_PP=0., 
                           XMgCO3=0., XCaCO3=0., DO=0., SH2=0., SCH4=0.):
        
           
        cmps = Components.load_default(default_compile=True)
        bst.settings.set_thermo(cmps)
        
        cmp_dct = dict.fromkeys(cmps.IDs, 0.)

        new = cls(ID=ID, phase=phase, T=T, P=P, units='kg/hr', price=price, 
                  thermo=thermo, pH=pH, SAlk=SAlk)

        if ratios: new.ratios = ratios
        else: new.ratios = WasteStream._default_ratios
        r = new._ratios

        #************ user-defined inorganic states **************        
        cmp_dct['SH2'] = SH2
        cmp_dct['SCH4'] = SCH4
        cmp_dct['SN2'] = SN2
        cmp_dct['SO2'] = DO
        cmp_dct['SNH4'] = SNH4
        cmp_dct['SNO2'] = SNO2
        cmp_dct['SNO3'] = SNO3
        cmp_dct['SPO4'] = SPO4
        cmp_dct['SCO3'] = SAlk * 12 * conc_unit.conversion_factor(units[1])       # 1 meq/L SAlk ~ 1 mmol/L HCO3- ~ 12 mg C/L (12 mg C/mmol HCO3-)
        cmp_dct['SCa'] = SCa
        cmp_dct['SMg'] = SMg
        cmp_dct['SK'] = SK
        cmp_dct['XMAP'] = XMAP
        cmp_dct['XHAP'] = XHAP
        cmp_dct['XHDP'] = XHDP
        cmp_dct['XFePO4'] = XFePO4
        cmp_dct['XAlPO4'] = XAlPO4
        cmp_dct['XFeOH'] = XFeOH
        cmp_dct['XAlOH'] = XAlOH
        cmp_dct['XMgCO3'] = XMgCO3
        cmp_dct['XCaCO3'] = XCaCO3
        cmp_dct['SCAT'] = SCAT
        cmp_dct['SAN'] = SAN
        
        #************ organic components **************
        sCOD = COD * iSCOD_COD
        sBOD = sCOD * iSBOD_SCOD
        cmp_dct['SCH3OH'] = SCH3OH
        cmp_dct['SAc'] = SAc
        cmp_dct['SProp'] = SProp
        
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        other_sBOD = (cmp_c * cmps.s * cmps.org * cmps.f_BOD5_COD).sum()
        cmp_dct['SF'] = (sBOD - other_sBOD)/cmps.SF.f_BOD5_COD
        
        cmp_c = np.asarray([v for v in cmp_dct.values()])        
        SU = sCOD - (cmp_c * cmps.s * cmps.org).sum()
        cmp_dct['SU_Inf'] = SU * r['iSUInf_SU']
        cmp_dct['SU_E'] = SU - cmp_dct['SU_Inf']
        
        cmp_dct['CB_BAP'] = CB * r['iBAP_CB']
        cmp_dct['CB_UAP'] = CB * r['iUAP_CB']
        cmp_dct['CB_Subst'] = CB - cmp_dct['CB_BAP'] - cmp_dct['CB_UAP']

        XCB = CB/r['iCB_XCB']
        XCU = COD - sCOD - XCB
        cmp_dct['XU_OHO_E'] = XU_OHO_E
        cmp_dct['XU_PAO_E'] = XU_PAO_E
        
        XCU_Inf = XCU - XU_OHO_E - XU_PAO_E
        cmp_dct['CU_Inf'] = XCU_Inf * r['iCUInf_XCUInf']
        cmp_dct['XU_Inf'] = XCU_Inf * (1 - r['iCUInf_XCUInf'])
        
        cmp_dct['XOHO'] = XOHO
        cmp_dct['XAOO'] = XAOO
        cmp_dct['XNOO'] = XNOO
        cmp_dct['XAMO'] = XAMO
        cmp_dct['XPAO'] = XPAO
        cmp_dct['XACO'] = XACO
        cmp_dct['XHMO'] = XHMO
        cmp_dct['XPRO'] = XPRO
        cmp_dct['XMEOLO'] = XMEOLO
        cmp_dct['XFO'] = XFO
        cmp_dct['XOHO_PHA'] = XOHO_PHA
        cmp_dct['XGAO_PHA'] = XGAO_PHA
        cmp_dct['XPAO_PHA'] = XPAO_PHA
        cmp_dct['XGAO_Gly'] = XGAO_Gly
        cmp_dct['XPAO_Gly'] = XPAO_Gly
                
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        cmp_dct['XB_Subst'] = COD - (cmp_c * cmps.org * (cmps.s + cmps.c + cmps.x)).sum()
        
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        VSS = (cmp_c * cmps.i_mass * cmps.f_Vmass_Totmass * cmps.x * cmps.org).sum()
        TSS = VSS/iVSS_TSS
        XOrg_ISS = (cmp_c * cmps.i_mass * (1-cmps.f_Vmass_Totmass) * cmps.x * cmps.org).sum()

        del sCOD, sBOD, other_sBOD, SU, XCU, XCB, XCU_Inf
        
        #************ inorganic components **************
        cmp_dct['XPAO_PP_Hi'] = XPAO_PP * r['iHi_XPAOPP']
        cmp_dct['XPAO_PP_Lo'] = XPAO_PP * (1 - r['iHi_XPAOPP'])
        
        ISS = TSS - VSS
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        other_ig_iss = (cmp_c * cmps.i_mass * cmps.x * (1-cmps.org)).sum()
        cmp_dct['XIg_ISS'] = ISS - XOrg_ISS - other_ig_iss
        
        del ISS, VSS, TSS, XOrg_ISS, other_ig_iss, cmp_c
        
        # TODO: calibrate pH, SAlk, SCAT, SAN
        bad_vars = {k:v for k,v in cmp_dct.items() if v<0}
        if len(bad_vars) > 0:
            raise ValueError(f"The following state variable(s) was found negative: {bad_vars}.")
        
        del bad_vars
        
        #************ calibrate XB_subst, SF's N, P content *************
        if SNH4 > 0 and cmp_dct['SF'] > 0:
            SFi_N = _calib_SF_iN(cmps, cmp_dct, SNH4/iSNH_STKN)
        XB_Substi_N = _calib_XBsub_iN(cmps, cmp_dct, TKN - SNH4/iSNH_STKN)
        XB_Substi_P = _calib_XBsub_iP(cmps, cmp_dct, TP)
        
        BOD = COD * iBOD_COD
        sub_IDs = ('XB_Subst', 'XOHO_PHA', 'XGAO_PHA', 'XPAO_PHA', 'XGAO_Gly', 'XPAO_Gly')
        fbodtocod_sub = _calib_XBsub_fBODCOD(cmps, cmp_dct, sub_IDs, BOD)
        
        #************ convert concentrations to flow rates *************
        flow_tot /= vol_unit.conversion_factor(units[0])
        factor = conc_unit.conversion_factor(units[1])
            
        cmp_dct = {k:v/factor*flow_tot*1e-6 for k,v in cmp_dct.items()}       # [mg/L]*[L/hr]*1e-6[kg/mg] = [kg/hr]
        dwt = sum(cmp_dct.values())
        
        den = 1
        i = 0
        while True:
            den0 = den
            cmp_dct['H2O'] = flow_tot*den0 - dwt
            new = cls(ID=ID, phase=phase, T=T, P=P, units='kg/hr', price=price, 
                      thermo=thermo, pH=pH, SAlk=SAlk, **cmp_dct)
            den = flow_tot*den0/(new.F_vol*1e3)            
            i += 1
            if abs(den-den0) <= 1e-3: break
            if i > 50: raise ValueError('Density calculation failed to converge within 50 iterations.')

        new.components.SF.i_N = SFi_N
        new.components.XB_Subst.i_N = XB_Substi_N
        new.components.XB_Subst.i_P = XB_Substi_P
        for i in sub_IDs: new.components[i].f_BOD5_COD = fbodtocod_sub
        
        return new


    @classmethod
    def bodbased_inf_model(cls, ID, flow_tot=0., units = ('L/hr', 'mg/L'), 
                           phase='l', T=298.15, P=101325., price=0., thermo=None, 
                           pH=7., SAlk=10., ratios=None, 
                           BOD=250., TKN=40., TP=10., iVSS_TSS=0.75, iSNH_STKN=0.9,
                           iSBOD_BOD=0.25, iSBOD_SCOD=0.50, iBOD_COD=0.58,
                           SN2=18., SNH4=25., SNO2=0., SNO3=0., SPO4=8., 
                           SCa=140., SMg=50., SK=28., SCAT=3., SAN=12.,  
                           CB=40., SCH3OH=0., SAc=0., SProp=0., 
                           XOHO_PHA=0., XGAO_PHA=0., XPAO_PHA=0., 
                           XGAO_Gly=0., XPAO_Gly=0., XU_OHO_E=0., XU_PAO_E=0.,
                           XOHO=0., XAOO=0., XNOO=0., XAMO=0., XPAO=0., 
                           XPRO=0., XACO=0., XHMO=0., XMEOLO=0., XFO=0.,
                           XFePO4=0., XAlPO4=0., XFeOH=0., XAlOH=0., 
                           XMAP=0., XHAP=0., XHDP=0., XPAO_PP=0., 
                           XMgCO3=0., XCaCO3=0., DO=0., SH2=0., SCH4=0.):
        
           
        cmps = Components.load_default(default_compile=True)
        bst.settings.set_thermo(cmps)
        
        cmp_dct = dict.fromkeys(cmps.IDs, 0.)

        new = cls(ID=ID, phase=phase, T=T, P=P, units='kg/hr', price=price, 
                  thermo=thermo, pH=pH, SAlk=SAlk)

        if ratios: new.ratios = ratios
        else: new.ratios = WasteStream._default_ratios
        r = new._ratios

        #************ user-defined inorganic states **************        
        cmp_dct['SH2'] = SH2
        cmp_dct['SCH4'] = SCH4
        cmp_dct['SN2'] = SN2
        cmp_dct['SO2'] = DO
        cmp_dct['SNH4'] = SNH4
        cmp_dct['SNO2'] = SNO2
        cmp_dct['SNO3'] = SNO3
        cmp_dct['SPO4'] = SPO4
        cmp_dct['SCO3'] = SAlk * 12 * conc_unit.conversion_factor(units[1])       # 1 meq/L SAlk ~ 1 mmol/L HCO3- ~ 12 mg C/L (12 mg C/mmol HCO3-)
        cmp_dct['SCa'] = SCa
        cmp_dct['SMg'] = SMg
        cmp_dct['SK'] = SK
        cmp_dct['XMAP'] = XMAP
        cmp_dct['XHAP'] = XHAP
        cmp_dct['XHDP'] = XHDP
        cmp_dct['XFePO4'] = XFePO4
        cmp_dct['XAlPO4'] = XAlPO4
        cmp_dct['XFeOH'] = XFeOH
        cmp_dct['XAlOH'] = XAlOH
        cmp_dct['XMgCO3'] = XMgCO3
        cmp_dct['XCaCO3'] = XCaCO3
        cmp_dct['SCAT'] = SCAT
        cmp_dct['SAN'] = SAN
        
        #************ organic components **************
        COD = BOD / iBOD_COD
        sBOD = BOD * iSBOD_BOD
        cmp_dct['SCH3OH'] = SCH3OH
        cmp_dct['SAc'] = SAc
        cmp_dct['SProp'] = SProp
        
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        other_sBOD = (cmp_c * cmps.s * cmps.org * cmps.f_BOD5_COD).sum()
        cmp_dct['SF'] = (sBOD - other_sBOD)/cmps.SF.f_BOD5_COD
        
        sCOD = sBOD / iSBOD_SCOD
        cmp_c = np.asarray([v for v in cmp_dct.values()])        
        SU = sCOD - (cmp_c * cmps.s * cmps.org).sum()
        cmp_dct['SU_Inf'] = SU * r['iSUInf_SU']
        cmp_dct['SU_E'] = SU - cmp_dct['SU_Inf']
        
        cmp_dct['CB_BAP'] = CB * r['iBAP_CB']
        cmp_dct['CB_UAP'] = CB * r['iUAP_CB']
        cmp_dct['CB_Subst'] = CB - cmp_dct['CB_BAP'] - cmp_dct['CB_UAP']

        XCB = CB/r['iCB_XCB']
        XCU = COD - sCOD - XCB
        cmp_dct['XU_OHO_E'] = XU_OHO_E
        cmp_dct['XU_PAO_E'] = XU_PAO_E
        
        XCU_Inf = XCU - XU_OHO_E - XU_PAO_E
        cmp_dct['CU_Inf'] = XCU_Inf * r['iCUInf_XCUInf']
        cmp_dct['XU_Inf'] = XCU_Inf * (1 - r['iCUInf_XCUInf'])
        
        cmp_dct['XOHO'] = XOHO
        cmp_dct['XAOO'] = XAOO
        cmp_dct['XNOO'] = XNOO
        cmp_dct['XAMO'] = XAMO
        cmp_dct['XPAO'] = XPAO
        cmp_dct['XACO'] = XACO
        cmp_dct['XHMO'] = XHMO
        cmp_dct['XPRO'] = XPRO
        cmp_dct['XMEOLO'] = XMEOLO
        cmp_dct['XFO'] = XFO
        cmp_dct['XOHO_PHA'] = XOHO_PHA
        cmp_dct['XGAO_PHA'] = XGAO_PHA
        cmp_dct['XPAO_PHA'] = XPAO_PHA
        cmp_dct['XGAO_Gly'] = XGAO_Gly
        cmp_dct['XPAO_Gly'] = XPAO_Gly
                
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        cmp_dct['XB_Subst'] = COD - (cmp_c * cmps.org * (cmps.s + cmps.c + cmps.x)).sum()
        
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        VSS = (cmp_c * cmps.i_mass * cmps.f_Vmass_Totmass * cmps.x * cmps.org).sum()
        TSS = VSS/iVSS_TSS
        XOrg_ISS = (cmp_c * cmps.i_mass * (1-cmps.f_Vmass_Totmass) * cmps.x * cmps.org).sum()

        del sCOD, sBOD, other_sBOD, SU, XCU, XCB, XCU_Inf
        
        #************ inorganic components **************
        cmp_dct['XPAO_PP_Hi'] = XPAO_PP * r['iHi_XPAOPP']
        cmp_dct['XPAO_PP_Lo'] = XPAO_PP * (1 - r['iHi_XPAOPP'])
        
        ISS = TSS - VSS
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        other_ig_iss = (cmp_c * cmps.i_mass * cmps.x * (1-cmps.org)).sum()
        cmp_dct['XIg_ISS'] = ISS - XOrg_ISS - other_ig_iss
        
        del ISS, VSS, TSS, XOrg_ISS, other_ig_iss, cmp_c
        
        # TODO: calibrate pH, SAlk, SCAT, SAN
        bad_vars = {k:v for k,v in cmp_dct.items() if v<0}
        if len(bad_vars) > 0:
            raise ValueError(f"The following state variable(s) was found negative: {bad_vars}.")
        
        del bad_vars
        
        #************ calibrate XB_subst, SF's N, P content *************
        if SNH4 > 0 and cmp_dct['SF'] > 0:
            SFi_N = _calib_SF_iN(cmps, cmp_dct, SNH4/iSNH_STKN)
        XB_Substi_N = _calib_XBsub_iN(cmps, cmp_dct, TKN - SNH4/iSNH_STKN)
        XB_Substi_P = _calib_XBsub_iP(cmps, cmp_dct, TP)
        
        BOD = COD * iBOD_COD
        sub_IDs = ('XB_Subst', 'XOHO_PHA', 'XGAO_PHA', 'XPAO_PHA', 'XGAO_Gly', 'XPAO_Gly')
        fbodtocod_sub = _calib_XBsub_fBODCOD(cmps, cmp_dct, sub_IDs, BOD)
        
        #************ convert concentrations to flow rates *************
        flow_tot /= vol_unit.conversion_factor(units[0])
        factor = conc_unit.conversion_factor(units[1])
            
        cmp_dct = {k:v/factor*flow_tot*1e-6 for k,v in cmp_dct.items()}       # [mg/L]*[L/hr]*1e-6[kg/mg] = [kg/hr]
        dwt = sum(cmp_dct.values())
        
        den = 1
        i = 0
        while True:
            den0 = den
            cmp_dct['H2O'] = flow_tot*den0 - dwt
            new = cls(ID=ID, phase=phase, T=T, P=P, units='kg/hr', price=price, 
                      thermo=thermo, pH=pH, SAlk=SAlk, **cmp_dct)
            den = flow_tot*den0/(new.F_vol*1e3)            
            i += 1
            if abs(den-den0) <= 1e-3: break
            if i > 50: raise ValueError('Density calculation failed to converge within 50 iterations.')
        
        new.components.SF.i_N = SFi_N
        new.components.XB_Subst.i_N = XB_Substi_N
        new.components.XB_Subst.i_P = XB_Substi_P        
        for i in sub_IDs: new.components[i].f_BOD5_COD = fbodtocod_sub

        return new


    @classmethod
    def sludge_inf_model(cls, ID, flow_tot=0., units = ('L/hr', 'mg/L'), 
                         phase='l', T=298.15, P=101325., price=0., thermo=None, 
                         pH=7., SAlk=10., ratios=None, 
                         TSS=1e4, TKN=750., TP=250., SNH4=100., SPO4=50., 
                         iVSS_TSS=0.65, iscCOD_COD=0.01, iSNH_STKN=0.9,
                         frXUInf_VSS=0.4, frXUE_VSS=0.3, frXOHO_VSS=0.2,  
                         frXAOO_VSS=0.01, frXNOO_VSS=0.01,frXPAO_VSS=0.01, 
                         frCB_scCOD=0.1, frSU_scCOD=0.8, 
                         SCa=140., SMg=50., SK=28., SCAT=3., SAN=12., SN2=18.,
                         frXACO_VSS=0., frXHMO_VSS=0., frXPRO_VSS=0., frXFO_VSS=0.,
                         frXMEOLO_VSS=0., frXAMO_VSS=0., 
                         frXOHO_PHA_VSS=0., frXGAO_PHA_VSS=0., frXPAO_PHA_VSS=0., 
                         frXGAO_Gly_VSS=0., frXPAO_Gly_VSS=0., 
                         frSCH3OH_scCOD=0., frSAc_scCOD=0., frSProp_scCOD=0.,
                         SNO2=0., SNO3=0., XPAO_PP=0., XFeOH=0., XAlOH=0., 
                         XFePO4=0., XAlPO4=0., XMAP=0., XHAP=0., XHDP=0., 
                         XMgCO3=0., XCaCO3=0., DO=0., SH2=0., SCH4=0.):
        
        
        cmps = Components.load_default(default_compile=True)
        bst.settings.set_thermo(cmps)
        
        cmp_dct = dict.fromkeys(cmps.IDs, 0.)

        new = cls(ID=ID, phase=phase, T=T, P=P, units='kg/hr', price=price, 
                  thermo=thermo, pH=pH, SAlk=SAlk)

        if ratios: new.ratios = ratios
        else: new.ratios = WasteStream._default_ratios
        r = new._ratios

        #************ user-defined inorganic states **************        
        cmp_dct['SH2'] = SH2
        cmp_dct['SCH4'] = SCH4
        cmp_dct['SN2'] = SN2
        cmp_dct['SO2'] = DO
        cmp_dct['SNH4'] = SNH4
        cmp_dct['SNO2'] = SNO2
        cmp_dct['SNO3'] = SNO3
        cmp_dct['SPO4'] = SPO4
        cmp_dct['SCO3'] = SAlk * 12 * conc_unit.conversion_factor(units[1])       # 1 meq/L SAlk ~ 1 mmol/L HCO3- ~ 12 mg C/L (12 mg C/mmol HCO3-)
        cmp_dct['SCa'] = SCa
        cmp_dct['SMg'] = SMg
        cmp_dct['SK'] = SK
        cmp_dct['XMAP'] = XMAP
        cmp_dct['XHAP'] = XHAP
        cmp_dct['XHDP'] = XHDP
        cmp_dct['XFePO4'] = XFePO4
        cmp_dct['XAlPO4'] = XAlPO4
        cmp_dct['XFeOH'] = XFeOH
        cmp_dct['XAlOH'] = XAlOH
        cmp_dct['XMgCO3'] = XMgCO3
        cmp_dct['XCaCO3'] = XCaCO3
        cmp_dct['SCAT'] = SCAT
        cmp_dct['SAN'] = SAN
        
        #************ particulate components **************
        VSS = TSS * iVSS_TSS
        cmp_dct['XU_Inf'] = frXUInf_VSS * VSS
        
        if r['iXUOHOE_XUE']: frOHO = r['iXUOHOE_XUE']
        else: 
            try: frOHO = frXOHO_VSS/(frXOHO_VSS + frXPAO_VSS)
            except ZeroDivisionError: frOHO = 0.5
        cmp_dct['XU_OHO_E'] = frXUE_VSS * VSS * frOHO
        cmp_dct['XU_PAO_E'] = frXUE_VSS * VSS * (1-frOHO)
        
        cmp_dct['XOHO'] = frXOHO_VSS * VSS
        cmp_dct['XAOO'] = frXAOO_VSS * VSS
        cmp_dct['XNOO'] = frXNOO_VSS * VSS
        cmp_dct['XAMO'] = frXAMO_VSS * VSS
        cmp_dct['XPAO'] = frXPAO_VSS * VSS
        cmp_dct['XACO'] = frXACO_VSS * VSS
        cmp_dct['XHMO'] = frXHMO_VSS * VSS
        cmp_dct['XPRO'] = frXPRO_VSS * VSS
        cmp_dct['XMEOLO'] = frXMEOLO_VSS * VSS
        cmp_dct['XFO'] = frXFO_VSS * VSS
        cmp_dct['XOHO_PHA'] = frXOHO_PHA_VSS * VSS
        cmp_dct['XGAO_PHA'] = frXGAO_PHA_VSS * VSS
        cmp_dct['XPAO_PHA'] = frXPAO_PHA_VSS * VSS
        cmp_dct['XGAO_Gly'] = frXGAO_Gly_VSS * VSS
        cmp_dct['XPAO_Gly'] = frXPAO_Gly_VSS * VSS
        
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        cmp_dct['XB_Subst'] = VSS - (cmp_c * cmps.org * cmps.x).sum()
        
        # convert gVSS to gCOD
        for cmp in cmps:
            if cmp.organic and cmp.particle_size == 'Particulate':
                cmp_dct[cmp.ID] /= cmp.i_mass * cmp.f_Vmass_Totmass
        
        cmp_dct['XPAO_PP_Hi'] = XPAO_PP * r['iHi_XPAOPP']
        cmp_dct['XPAO_PP_Lo'] = XPAO_PP * (1 - r['iHi_XPAOPP'])
        
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        ig_ISS = TSS - (cmp_c * cmps.i_mass * cmps.x * cmps.org).sum()
        other_ig_iss = (cmp_c * cmps.i_mass * cmps.x * (1-cmps.org)).sum()
        cmp_dct['XIg_ISS'] = ig_ISS - other_ig_iss
        
        del other_ig_iss, cmp_c
        
        #*********** soluble and colloidal components *************
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        xCOD = (cmp_c * cmps.x * cmps.org).sum()
        scCOD = xCOD * iscCOD_COD / (1-iscCOD_COD)
        
        cmp_dct['SCH3OH'] = frSCH3OH_scCOD * scCOD
        cmp_dct['SAc'] = frSAc_scCOD * scCOD
        cmp_dct['SProp'] = frSProp_scCOD * scCOD
        cmp_dct['SU_Inf'] = frSU_scCOD * scCOD * r['iSUInf_SU']
        cmp_dct['SU_E'] = frSU_scCOD * scCOD - cmp_dct['SU_Inf']
        
        CB = frCB_scCOD * scCOD
        cmp_dct['CB_BAP'] = CB * r['iBAP_CB']
        cmp_dct['CB_UAP'] = CB * r['iUAP_CB']
        cmp_dct['CB_Subst'] = CB - cmp_dct['CB_BAP'] - cmp_dct['CB_UAP']        
        cmp_dct['CU_Inf'] = cmp_dct['XU_Inf'] * r['iCUInf_XCUInf'] / (1-r['iCUInf_XCUInf'])
        
        cmp_c = np.asarray([v for v in cmp_dct.values()])
        cmp_dct['SF'] = scCOD - (cmp_c * (cmps.s + cmps.c) * cmps.org).sum()
                
        # TODO: calibrate pH, SAlk, SCAT, SAN
        bad_vars = {k:v for k,v in cmp_dct.items() if v<0}
        if len(bad_vars) > 0:
            raise ValueError(f"The following state variable(s) was found negative: {bad_vars}.")
        
        del bad_vars
        
        #************ calibrate XB_subst, SF's N, P content *************
        if SNH4 > 0 and cmp_dct['SF'] > 0:
            SFi_N = _calib_SF_iN(cmps, cmp_dct, SNH4/iSNH_STKN)
        XB_Substi_N = _calib_XBsub_iN(cmps, cmp_dct, TKN - SNH4/iSNH_STKN)
        XB_Substi_P = _calib_XBsub_iP(cmps, cmp_dct, TP)

        #************ convert concentrations to flow rates *************
        flow_tot /= vol_unit.conversion_factor(units[0])
        factor = conc_unit.conversion_factor(units[1])
            
        cmp_dct = {k:v/factor*flow_tot*1e-6 for k,v in cmp_dct.items()}       # [mg/L]*[L/hr]*1e-6[kg/mg] = [kg/hr]
        dwt = sum(cmp_dct.values())
        
        den = 1
        i = 0
        while True:
            den0 = den
            cmp_dct['H2O'] = flow_tot*den0 - dwt
            new = cls(ID=ID, phase=phase, T=T, P=P, units='kg/hr', price=price, 
                      thermo=thermo, pH=pH, SAlk=SAlk, **cmp_dct)
            den = flow_tot*den0/(new.F_vol*1e3)            
            i += 1
            if abs(den-den0) <= 1e-3: break
            if i > 50: raise ValueError('Density calculation failed to converge within 50 iterations.')
        

        new.components.SF.i_N = SFi_N
        new.components.XB_Subst.i_N = XB_Substi_N
        new.components.XB_Subst.i_P = XB_Substi_P        

        return new

#%% functions for calibrations of N, P contents and BOD:COD ratio of certain components
        
def _calib_SF_iN(components, concentrations, STKN):    
    cmp_c = np.asarray([v for v in concentrations.values()])
    SN = (cmp_c * components.i_N * (components.s + components.c)).sum()
    SF_N = concentrations['SF'] * components.SF.i_N
    SNOx_N = concentrations['SNO2'] * components.SNO2.i_N + concentrations['SNO3'] * components.SNO3.i_N
    other_stkn = SN - SF_N - SNOx_N
    SF_N = STKN - other_stkn           
    if SF_N < 0:
        raise ValueError("Negative N content for SF was estimated.")                        
    return SF_N/concentrations['SF']
    
def _calib_XBsub_iN(components, concentrations, XTKN):
    cmp_c = np.asarray([v for v in concentrations.values()])        
    other_xtkn = (cmp_c * components.i_N * components.x).sum() - concentrations['XB_Subst'] * components.XB_Subst.i_N                
    XB_Subst_N = XTKN - other_xtkn
    if XB_Subst_N < 0:
        raise ValueError("Negative N content for XB_Subst was estimated.")            
    return XB_Subst_N/concentrations['XB_Subst']
    

def _calib_XBsub_iP(components, concentrations, TP):
    cmp_c = np.asarray([v for v in concentrations.values()])    
    other_p = (cmp_c * components.i_P).sum() - concentrations['XB_Subst'] * components.XB_Subst.i_P
    XB_Subst_P = TP - other_p
    if XB_Subst_P < 0:
        raise ValueError("Negative P content for XB_Subst was estimated.")    
    return XB_Subst_P/concentrations['XB_Subst']
    
def _calib_XBsub_fBODCOD(components, concentrations, substrate_IDs, BOD):
    cmp_c = np.asarray([v for v in concentrations.values()])
    c_sub = np.asarray([v for k,v in concentrations.items() if k in substrate_IDs])
    XB_sub = components.subgroup(substrate_IDs)
    other_BOD = (cmp_c * (components.x + components.c + components.s) * components.f_BOD5_COD).sum() - (c_sub * XB_sub.f_BOD5_COD).sum()
    fbodtocod_sub = (BOD - other_BOD)/c_sub.sum()
    if fbodtocod_sub > 1 or fbodtocod_sub < 0:
        raise ValueError("BOD5-to-COD ratio for XB_Subst and XStor was estimated out of range [0,1].")
    return fbodtocod_sub

